import os

from .cache import *  # will also import some globals like `britive`


def test_create(cached_profile):
    assert isinstance(cached_profile, dict)
    assert cached_profile['name'].startswith('test')
    assert len(cached_profile['scope']) == 0


def test_list(cached_profile):
    profiles = britive.profiles.list(application_id=cached_profile['appContainerId'])
    assert isinstance(profiles, list)
    assert len(profiles) > 0
    assert isinstance(profiles[0], dict)
    assert profiles[0]['name'].startswith('test')


def test_list_include_policies(cached_profile):
    profiles = britive.profiles.list(application_id=cached_profile['appContainerId'], include_policies=True)
    assert isinstance(profiles, list)
    assert len(profiles) > 0
    assert isinstance(profiles[0], dict)
    assert profiles[0]['profileName'].startswith('test')
    assert 'policies' in profiles[0]


def test_get(cached_profile):
    profile = britive.profiles.get(application_id=cached_profile['appContainerId'], profile_id=cached_profile['papId'])
    assert isinstance(profile, dict)
    assert profile['name'].startswith('test')


def test_update(cached_profile):
    profile = britive.profiles.update(
        application_id=cached_profile['appContainerId'],
        profile_id=cached_profile['papId'],
        description='test desc'
    )
    assert isinstance(profile, dict)
    assert profile['description'] == 'test desc'


def test_set_scopes(cached_profile, cached_environment):
    scopes = britive.profiles.set_scopes(
        profile_id=cached_profile['papId'],
        scopes=[
            {
                'type': 'Environment',
                'value': cached_environment['id']
            }
        ]
    )
    assert isinstance(scopes, list)
    assert len(scopes) == 1
    assert isinstance(scopes[0], dict)
    assert scopes[0]['value'] == cached_environment['id']


def test_get_scopes(cached_profile, cached_environment):
    scopes = britive.profiles.get_scopes(profile_id=cached_profile['papId'])
    assert isinstance(scopes, list)
    assert len(scopes) == 1
    assert isinstance(scopes[0], dict)
    assert scopes[0]['value'] == cached_environment['id']


def test_remove_single_environment_scope(cached_profile, cached_environment):
    response = britive.profiles.remove_single_environment_scope(
        profile_id=cached_profile['papId'],
        environment_id=cached_environment['id']
    )
    assert response is None


def test_add_single_environment_scope(cached_profile, cached_environment):
    response = britive.profiles.add_single_environment_scope(
        profile_id=cached_profile['papId'],
        environment_id=cached_environment['id']
    )
    assert response is None


def test_disable(cached_profile):
    profile = britive.profiles.disable(
        application_id=cached_profile['appContainerId'],
        profile_id=cached_profile['papId']
    )
    assert isinstance(profile, dict)
    assert profile['status'] == 'inactive'


def test_enable(cached_profile):
    profile = britive.profiles.enable(
        application_id=cached_profile['appContainerId'],
        profile_id=cached_profile['papId']
    )
    assert isinstance(profile, dict)
    assert profile['status'] == 'active'


def test_session_attributes_add_static(cached_static_session_attribute):
    assert isinstance(cached_static_session_attribute, dict)


def test_session_attributes_add_dynamic(cached_dynamic_session_attribute):
    assert isinstance(cached_dynamic_session_attribute, dict)


def test_session_attributes_list(cached_profile):
    attributes = britive.profiles.session_attributes.list(profile_id=cached_profile['papId'])
    assert isinstance(attributes, list)
    assert len(attributes) == 2


def test_session_attributes_update_static(cached_profile, cached_static_session_attribute):
    attribute = britive.profiles.session_attributes.update_static(
        profile_id=cached_profile['papId'],
        attribute_id=cached_static_session_attribute['id'],
        tag_name='test-static-2',
        tag_value='test',
        transitive=True
    )
    assert attribute is None


def test_session_attributes_update_dynamic(cached_profile, cached_dynamic_session_attribute):
    attribute = britive.profiles.session_attributes.update_dynamic(
        profile_id=cached_profile['papId'],
        attribute_id=cached_dynamic_session_attribute['id'],
        identity_attribute_id='w2zQ4R9xoyrkWY4phEV9',
        tag_name='test-dynamic-2',
        transitive=True
    )
    assert attribute is None


def test_session_attributes_remove(cached_profile, cached_static_session_attribute, cached_dynamic_session_attribute):
    try:
        static = britive.profiles.session_attributes.remove(
            profile_id=cached_profile['papId'],
            attribute_id=cached_static_session_attribute['id']
        )
        assert static is None

        dynamic = britive.profiles.session_attributes.remove(
            profile_id=cached_profile['papId'],
            attribute_id=cached_dynamic_session_attribute['id']
        )
        assert dynamic is None
    finally:
        for resource in ['static', 'dynamic']:
            file = f'./.cache/v/resources/{resource}-session-attribute'
            if os.path.isfile(file):
                os.remove(file)


def test_policies_create(cached_profile_policy):
    assert isinstance(cached_profile_policy, dict)
    assert cached_profile_policy['members']['tags']


def test_policies_create_with_approval_single_notification_medium(cached_profile_approval_policy):
    assert isinstance(cached_profile_approval_policy, dict)
    assert cached_profile_approval_policy['members']['serviceIdentities']


def test_policies_create_with_approval_multiple_notification_medium(cached_profile, cached_service_identity):
    policy = britive.profiles.policies.build(
        name=f"{cached_profile['papId']}-2",
        description='',
        service_identities=[cached_service_identity['username']],
        approval_notification_medium=['Email'],
        approver_users=[britive.my_access.whoami()['username']]
    )
    assert isinstance(policy, dict)


def test_policies_list(cached_profile):
    policies = britive.profiles.policies.list(profile_id=cached_profile['papId'])
    assert isinstance(policies, list)


def test_policies_get(cached_profile, cached_profile_policy):
    policy = britive.profiles.policies.get(
        profile_id=cached_profile['papId'],
        policy_id=cached_profile_policy['id']
    )
    assert isinstance(policy, dict)


def test_policies_condition_created_as_json_get_formatted_json(cached_profile, cached_profile_policy_condition_as_json_str):
    policy = britive.profiles.policies.get(
        profile_id=cached_profile['papId'],
        policy_id=cached_profile_policy_condition_as_json_str['id'], condition_as_dict=False
    )
    assert isinstance(policy['condition'], str)


def test_policies_condition_created_as_json_get_formatted_dict(cached_profile, cached_profile_policy_condition_as_json_str):
    policy = britive.profiles.policies.get(
        profile_id=cached_profile['papId'],
        policy_id=cached_profile_policy_condition_as_json_str['id'],
        condition_as_dict=True
    )
    assert isinstance(policy['condition'], dict)


def test_policies_condition_created_as_dict_get_formatted_json(cached_profile, cached_profile_policy_condition_as_dict):
    policy = britive.profiles.policies.get(
        profile_id=cached_profile['papId'],
        policy_id=cached_profile_policy_condition_as_dict['id'], condition_as_dict=False
    )
    assert isinstance(policy['condition'], str)


def test_policies_condition_created_as_dict_get_formatted_dict(cached_profile, cached_profile_policy_condition_as_dict):
    policy = britive.profiles.policies.get(
        profile_id=cached_profile['papId'],
        policy_id=cached_profile_policy_condition_as_dict['id'], condition_as_dict=True
    )
    assert isinstance(policy['condition'], dict)


def test_policies_update(cached_profile, cached_profile_policy):
    policy = {
        'members': {
            'tags': [
                {'id': tag['id']}
                for tag in cached_profile_policy['members']['tags']
            ]
        }
    }
    assert britive.profiles.policies.update(
        profile_id=cached_profile['papId'],
        policy_id=cached_profile_policy['id'],
        policy=policy
    ) is None


def test_policies_delete(cached_profile, cached_profile_policy):
    try:
        assert britive.profiles.policies.delete(
            profile_id=cached_profile['papId'],
            policy_id=cached_profile_policy['id']
        ) is None
    finally:
        cleanup('profile-policy')


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_list_supported_types(cached_gcp_profile_big_query, cached_gcp_profile_storage):
    response = britive.profiles.permissions.constraints.list_supported_types(
        profile_id=cached_gcp_profile_big_query['papId'],
        permission_name='BigQuery Admin',
        permission_type='role'
    )
    assert 'bigquery.datasets' in response
    assert 'bigquery.tables' in response

    response = britive.profiles.permissions.constraints.list_supported_types(
        profile_id=cached_gcp_profile_storage['papId'],
        permission_name='Storage Admin',
        permission_type='role'
    )

    assert len(response) == 1
    assert 'condition' in response


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_get_before_add(cached_gcp_profile_big_query, cached_gcp_profile_storage):
    response = britive.profiles.permissions.constraints.get(
        profile_id=cached_gcp_profile_big_query['papId'],
        permission_name='BigQuery Admin',
        permission_type='role',
        constraint_type='bigquery.datasets'
    )

    assert response is None

    response = britive.profiles.permissions.constraints.get(
        profile_id=cached_gcp_profile_big_query['papId'],
        permission_name='BigQuery Admin',
        permission_type='role',
        constraint_type='bigquery.tables'
    )

    assert response is None

    response = britive.profiles.permissions.constraints.get(
        profile_id=cached_gcp_profile_storage['papId'],
        permission_name='Storage Admin',
        permission_type='role',
        constraint_type='condition'
    )

    assert response is None


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_lint_condition(cached_gcp_profile_storage):
    expression = "(resource.type != 'storage.googleapis.com/Bucket' && " \
                 "resource.type != 'storage.googleapis.com/Object') || " \
                 "resource.name.startsWith('projects/_/buckets/my-first-project-demo-bucket-1')"
    response = britive.profiles.permissions.constraints.lint_condition(
        profile_id=cached_gcp_profile_storage['papId'],
        permission_name='Storage Admin',
        permission_type='role',
        expression=expression
    )

    assert response['success']
    assert response['message'] == '{}'


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_add_big_query(cached_gcp_profile_big_query):
    response = britive.profiles.permissions.constraints.add(
        profile_id=cached_gcp_profile_big_query['papId'],
        permission_name='BigQuery Admin',
        permission_type='role',
        constraint_type='bigquery.datasets',
        constraint='my-first-project-310615.myfirstdataset'
    )

    assert response is None

    try:
        britive.profiles.permissions.constraints.add(
            profile_id=cached_gcp_profile_big_query['papId'],
            permission_name='BigQuery Admin',
            permission_type='role',
            constraint_type='bigquery.datasets',
            constraint='my-first-project-310615.myfirstdataset'
        )
    except Exception as e:
        assert 'Constraint is already available in the added list' in str(e)


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_add_storage(cached_gcp_profile_storage):
    expression = "(resource.type != 'storage.googleapis.com/Bucket' && " \
                 "resource.type != 'storage.googleapis.com/Object') || " \
                 "resource.name.startsWith('projects/_/buckets/my-first-project-demo-bucket-1')"

    constraint = {
        'title': 'test',
        'description': 'test',
        'expression': expression
    }

    response = britive.profiles.permissions.constraints.add(
        profile_id=cached_gcp_profile_storage['papId'],
        permission_name='Storage Admin',
        permission_type='role',
        constraint_type='condition',
        constraint=constraint
    )

    assert response is None

    try:
        britive.profiles.permissions.constraints.add(
            profile_id=cached_gcp_profile_storage['papId'],
            permission_name='Storage Admin',
            permission_type='role',
            constraint_type='condition',
            constraint=constraint
        )
    except Exception as e:
        assert 'Constraint is already available in the added list' in str(e)


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_remove_big_query(cached_gcp_profile_big_query):
    try:
        response = britive.profiles.permissions.constraints.remove(
            profile_id=cached_gcp_profile_big_query['papId'],
            permission_name='BigQuery Admin',
            permission_type='role',
            constraint_type='bigquery.datasets',
            constraint='my-first-project-310615.myfirstdataset'
        )

        assert response is None
    finally:
        britive.profiles.delete(
            application_id=os.getenv('BRITIVE_GCP_TEST_APP_ID'),
            profile_id=cached_gcp_profile_big_query['papId']
        )
        cleanup('gcp-profile-bq')


@pytest.mark.skipif(constraints, reason=constraints_skip)
def test_constraints_remove_storage(cached_gcp_profile_storage):
    try:
        response = britive.profiles.permissions.constraints.remove(
            profile_id=cached_gcp_profile_storage['papId'],
            permission_name='Storage Admin',
            permission_type='role',
            constraint_type='condition',
            constraint=None
        )

        assert response is None
    finally:
        britive.profiles.delete(
            application_id=os.getenv('BRITIVE_GCP_TEST_APP_ID'),
            profile_id=cached_gcp_profile_storage['papId']
        )
        cleanup('gcp-profile-storage')



from .cache import *  # will also import some globals like `britive`


def test_create(cached_security_policy):
    assert isinstance(cached_security_policy, dict)


def test_list(cached_security_policy):
    policies = britive.security_policies.list()
    assert isinstance(policies, list)
    assert cached_security_policy['id'] in [p['id'] for p in policies]


def test_get(cached_security_policy):
    policy = britive.security_policies.get(security_policy_id=cached_security_policy['id'])
    assert isinstance(policy, dict)


def test_disable(cached_security_policy):
    response = britive.security_policies.disable(security_policy_id=cached_security_policy['id'])
    assert response is None
    policy = britive.security_policies.get(security_policy_id=cached_security_policy['id'])
    assert policy['status'] == 'Inactive'


def test_enable(cached_security_policy):
    response = britive.security_policies.enable(security_policy_id=cached_security_policy['id'])
    assert response is None
    policy = britive.security_policies.get(security_policy_id=cached_security_policy['id'])
    assert policy['status'] == 'Active'


def test_update(cached_security_policy):
    response = britive.security_policies.update(
        security_policy_id=cached_security_policy['id'],
        ips=['2.2.2.2']
    )
    assert response is None

    policy = britive.security_policies.get(security_policy_id=cached_security_policy['id'])
    assert policy['conditions'][0]['values'] == ['2.2.2.2']


def test_delete(cached_security_policy):
    response = britive.security_policies.delete(security_policy_id=cached_security_policy['id'])
    assert response is None
    cleanup('security-policy')



import time
import britive.exceptions
from britive import policies
from .cache import *  # will also import some globals like `britive`


def test_create_vault(cached_vault):
    assert isinstance(cached_vault, dict)


def test_list_vault(cached_vault):
    response = britive.secrets_manager.vaults.list()
    assert isinstance(response, dict)
    assert len(response) > 0


def test_get_vault(cached_vault):
    vault = britive.secrets_manager.vaults.get_vault_by_id(cached_vault['id'])
    assert isinstance(vault, dict)
    assert vault['id'] == cached_vault['id']


def test_update_vault(cached_vault):
    britive.secrets_manager.vaults.update(cached_vault['id'], description='12345')
    vault = britive.secrets_manager.vaults.get_vault_by_id(cached_vault['id'])
    assert vault['description'] == '12345'


def test_create_folder(cached_folder):
    assert isinstance(cached_folder, dict)


def test_PasswordPolicies_create(cached_PasswordPolicies):
    assert isinstance(cached_PasswordPolicies, dict)


def test_PinPolicies_create(cached_PinPolicies):
    assert isinstance(cached_PinPolicies, dict)


def test_PasswordPolicies_get(cached_PasswordPolicies):
    pwdpolicy = britive.secrets_manager.password_policies.get(cached_PasswordPolicies['id'])
    assert isinstance(pwdpolicy, dict)
    assert pwdpolicy['id'] == cached_PasswordPolicies['id']


def test_PasswordPolicies_list():
    response = britive.secrets_manager.password_policies.list()
    assert isinstance(response, list)
    assert len(response) > 0
    assert isinstance(response[0], dict)


def test_PasswordPolicies_update(cached_PasswordPolicies):
    r = str(random.randint(0, 1000000))
    new_name = f"pytestpolicy-{r}"
    britive.secrets_manager.password_policies.update(cached_PasswordPolicies['id'], name=new_name)
    assert britive.secrets_manager.password_policies.get(
        cached_PasswordPolicies['id'])['name'] == new_name


def test_generate_password(cached_PasswordPolicies):
    time.sleep(5)  # sleep for 5 seconds as I think the password policy update takes a bit to settle down
    password = britive.secrets_manager.password_policies.generate_password(
        cached_PasswordPolicies['id'])
    assert isinstance(password, str)


def test_validate_password(cached_PasswordPolicies):
    response = britive.secrets_manager.password_policies.validate(
        cached_PasswordPolicies['id'],
        'testpassword')
    assert response['isPasswordOrPinValid'] is False


def test_static_secret_template_create(cached_static_secret_template):
    assert isinstance(cached_static_secret_template, dict)


def test_static_secret_template_list():
    response = britive.secrets_manager.static_secret_templates.list()
    assert isinstance(response, list)
    assert len(response)>0
    assert isinstance(response[0], dict)


def test_static_secret_template_get(cached_static_secret_template):
    static_secret_template = britive.secrets_manager.static_secret_templates.get(
        cached_static_secret_template['id'])
    assert isinstance(static_secret_template, dict)
    assert cached_static_secret_template['id'] == static_secret_template['id']


def test_static_secret_template_update(cached_static_secret_template):
    britive.secrets_manager.static_secret_templates.update(
        cached_static_secret_template['id'],
        description='test desc')
    assert britive.secrets_manager.static_secret_templates.get(
        cached_static_secret_template['id'])['description'] == 'test desc'


def test_secret_create(cached_secret):
    assert isinstance(cached_secret, dict)


def test_secret_get(cached_vault): 
    secret = britive.secrets_manager.secrets.get(
        vault_id=cached_vault['id'],
        path='/', secret_type='secret')[0]
    assert isinstance(secret, dict)


def test_secret_update(cached_secret, cached_vault):
    update = britive.secrets_manager.secrets.update(
        cached_vault['id'],
        cached_secret['path'],
        {'Note': 'updated test note'}
    )
    assert update is None

def test_secret_rename(cached_secret, cached_vault):
    name = cached_secret['name']
    rename = britive.secrets_manager.secrets.rename(
        cached_vault['id']
        , cached_secret['path']
        , 'testSecretRename')
    assert rename is None

    new_path_list = cached_secret['path'].split('/')
    new_path_list.pop()
    new_path_list.append('testSecretRename')
    new_path = '/'.join(new_path_list)

    rollback = britive.secrets_manager.secrets.rename(
        cached_vault['id']
        , new_path
        , name)
    assert rollback is None


def test_policies_create(cached_policy):
    assert isinstance(cached_policy, dict)


def test_policies_list():
    sec_policies = britive.secrets_manager.policies.list()
    assert isinstance(sec_policies, list)
    assert isinstance(sec_policies[0], dict)


def test_resources_get():
    response = britive.secrets_manager.resources.get('/')
    assert isinstance(response, dict)


def test_rotate_keys(cached_vault):
    initial_time = britive.secrets_manager.vaults.get_vault_by_id(cached_vault['id'])['lastRotation']
    britive.secrets_manager.vaults.rotate_keys()
    current_time = britive.secrets_manager.vaults.get_vault_by_id(cached_vault['id'])['lastRotation']
    assert initial_time != current_time
    tries = 0
    while True:
        try:
            if tries >= 30:
                raise Exception('timed out - vault rotation took too long')
            britive.my_secrets.list()
            break
        except exceptions.NoSecretsVaultFound:
            time.sleep(3)
            tries += 1
        except exceptions.NotFound:
            time.sleep(3)
            tries += 1





from .cache import *  # will also import some globals like `britive`


def test_create(cached_notification_medium):
    assert isinstance(cached_notification_medium, dict)
    assert 'pytest-nm-' in cached_notification_medium['name']


def test_list():
    response = britive.notification_mediums.list()
    assert isinstance(response, list)
    assert isinstance(response[0], dict)


def test_get(cached_notification_medium):
    response = britive.notification_mediums.get(cached_notification_medium['id'])
    assert isinstance(response, dict)
    assert response['name'] == cached_notification_medium['name']


def test_update(cached_notification_medium):
    r = str(random.randint(0, 1000000))
    new_name = f'pytest-nm-{r}'
    britive.notification_mediums.update(
        cached_notification_medium['id'],
        parameters={'name': new_name}
    )
    response = britive.notification_mediums.get(cached_notification_medium['id'])
    assert response['name'] == new_name



from .cache import *  # will also import some globals like `britive`


def test_environment_group_create(cached_environment_group):
    assert isinstance(cached_environment_group, dict)


def test_environment_group_list(cached_application):
    groups = britive.environment_groups.list(application_id=cached_application['appContainerId'])
    assert isinstance(groups, list)
    assert len(groups) > 0
    assert isinstance(groups[0], dict)


def test_environment_group_get(cached_application, cached_environment_group):
    group_get = britive.environment_groups.get(
        application_id=cached_application['appContainerId'], 
        environment_group_id=cached_environment_group['id']
    )
    assert isinstance(group_get, dict)
    assert group_get['name'] == cached_environment_group['name']








from .cache import *  # will also import some globals like `britive`
from britive import exceptions


def test_identity_provider_list():
    response = britive.workload.identity_providers.list()
    assert isinstance(response, list)


def test_identity_provider_create_aws(cached_workload_identity_provider_aws):
    # doing less work in the validation here (vs oidc) since an aws provider can already
    # exist and we are only allowed to create 1 per tenant
    assert isinstance(cached_workload_identity_provider_aws, dict)
    assert 'id' in cached_workload_identity_provider_aws.keys()


def test_identity_provider_create_oidc(cached_workload_identity_provider_oidc):
    assert isinstance(cached_workload_identity_provider_oidc, dict)
    assert 'id' in cached_workload_identity_provider_oidc.keys()
    assert cached_workload_identity_provider_oidc['name'].startswith('python-sdk-oidc')
    assert isinstance(cached_workload_identity_provider_oidc['attributesMap'], list)
    assert len(cached_workload_identity_provider_oidc['attributesMap']) == 1


def test_identity_provider_get(cached_workload_identity_provider_oidc):
    response = britive.workload.identity_providers.get(
        workload_identity_provider_id=cached_workload_identity_provider_oidc['id']
    )
    assert isinstance(response, dict)
    assert response['id'] == cached_workload_identity_provider_oidc['id']


def test_identity_provider_update_aws(cached_workload_identity_provider_aws):
    # we may very well need to set the update back to what it was originally as
    # we can only have 1 aws provider, so we want to restore that provider to its
    # original state
    existing_max_duration = britive.workload.identity_providers.get(
        workload_identity_provider_id=cached_workload_identity_provider_aws['id']
    )['maxDuration']

    # do the update
    response = britive.workload.identity_providers.update_aws(
        workload_identity_provider_id=cached_workload_identity_provider_aws['id'],
        max_duration=1
    )
    assert isinstance(response, dict)
    assert response['maxDuration'] == 1

    # restore it back
    britive.workload.identity_providers.update_aws(
        workload_identity_provider_id=cached_workload_identity_provider_aws['id'],
        max_duration=existing_max_duration
    )


def test_identity_provider_update_oidc(cached_workload_identity_provider_oidc):
    response = britive.workload.identity_providers.update_oidc(
        workload_identity_provider_id=cached_workload_identity_provider_oidc['id'],
        issuer_url='https://id2.fakedomain.com'
    )
    assert isinstance(response, dict)
    assert response['issuerUrl'] == 'https://id2.fakedomain.com'


def test_generate_attribute_map(cached_identity_attribute):
    response = britive.workload.identity_providers.generate_attribute_map(
        idp_attribute_name='sub',
        custom_identity_attribute_id=cached_identity_attribute['id']
    )
    assert isinstance(response, dict)
    assert 'idpAttr' in response.keys()
    assert 'userAttr' in response.keys()
    assert response['idpAttr'] == 'sub'
    assert response['userAttr'] == cached_identity_attribute['id']

    response = britive.workload.identity_providers.generate_attribute_map(
        idp_attribute_name='sub',
        custom_identity_attribute_name=cached_identity_attribute['name']
    )
    assert isinstance(response, dict)
    assert 'idpAttr' in response.keys()
    assert 'userAttr' in response.keys()
    assert response['idpAttr'] == 'sub'
    assert response['userAttr'] == cached_identity_attribute['id']


def test_service_identity_get_when_nothing_associated(cached_service_identity):
    with pytest.raises(exceptions.NotFound):
        britive.workload.service_identities.get(
            service_identity_id=cached_service_identity['userId']
        )


def test_service_identity_assign_and_unassign(cached_service_identity, cached_identity_attribute,
                                              cached_workload_identity_provider_oidc):
    response = britive.workload.service_identities.assign(
        service_identity_id=cached_service_identity['userId'],
        idp_id=cached_workload_identity_provider_oidc['id'],
        federated_attributes={
            cached_identity_attribute['id']: 'test'
        }
    )
    assert isinstance(response, dict)

    attrs = britive.service_identities.custom_attributes.get(
        principal_id=cached_service_identity['userId'],
        as_dict=True
    )

    assert isinstance(attrs, dict)
    assert len(attrs.keys()) == 1
    assert attrs[cached_identity_attribute['id']] == 'test'

    response = britive.workload.service_identities.unassign(
        service_identity_id=cached_service_identity['userId']
    )

    assert response is None

    attrs = britive.service_identities.custom_attributes.get(
        principal_id=cached_service_identity['userId'],
        as_dict=False
    )

    assert isinstance(attrs, list)
    assert len(attrs) == 0

    response = britive.workload.service_identities.assign(
        service_identity_id=cached_service_identity['userId'],
        idp_id=cached_workload_identity_provider_oidc['id'],
        federated_attributes={
            cached_identity_attribute['name']: 'test'
        }
    )
    assert isinstance(response, dict)

    attrs = britive.service_identities.custom_attributes.get(
        principal_id=cached_service_identity['userId'],
        as_dict=True
    )

    assert isinstance(attrs, dict)
    assert len(attrs.keys()) == 1
    assert attrs[cached_identity_attribute['id']] == 'test'

    response = britive.workload.service_identities.unassign(
        service_identity_id=cached_service_identity['userId']
    )

    assert response is None

    attrs = britive.service_identities.custom_attributes.get(
        principal_id=cached_service_identity['userId'],
        as_dict=False
    )

    assert isinstance(attrs, list)
    assert len(attrs) == 0


def test_identity_provider_delete(cached_workload_identity_provider_oidc, cached_workload_identity_provider_aws):
    try:
        # we do not want to delete the pre-existing aws provider
        if cached_workload_identity_provider_aws['name'].startswith('python-sdk-aws'):
            aws = britive.workload.identity_providers.delete(
                workload_identity_provider_id=cached_workload_identity_provider_aws['id']
            )
            assert aws is None
        oidc = britive.workload.identity_providers.delete(
            workload_identity_provider_id=cached_workload_identity_provider_oidc['id']
        )
        assert oidc is None
    finally:
        cleanup('workload-identity-provider-oidc')
        cleanup('workload-identity-provider-aws')








from .cache import *  # will also import some globals like `britive`


def test_list():
    response1 = britive.system.actions.list()
    assert isinstance(response1, list)
    assert len(response1) > 0
    assert isinstance(response1[0], dict)
    assert 'name' in response1[0].keys()
    assert 'tenantId' in response1[0].keys()

    response2 = britive.system.actions.list(consumer='apps')
    assert isinstance(response2, list)
    assert len(response2) > 0
    assert isinstance(response2[0], dict)
    assert 'name' in response2[0].keys()
    assert 'tenantId' in response2[0].keys()

    assert len(response2) < len(response1)

    consumers = list(set([c['consumer'] for c in response2]))

    assert len(consumers) == 1
    assert consumers[0] == 'apps'




valid_user_account_mappings = {
    'email': [
        {
            'name': 'email',
            'description': 'Email'
        }
    ],
    'username': [
        {
            'name': 'username',
            'description': 'Username'
        }
    ]
}


class Applications:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'

    def list(self, extended: bool = True) -> list:
        """
        Return a list of applications in the Britive tenant.

        :param extended: if True, will return additional details of the applications
        :return: List of applications.
        """

        params = {}
        if extended:
            params['view'] = 'extended'
        return self.britive.get(self.base_url, params=params)

    def get(self, application_id: str) -> dict:
        """
        Return details of the specified application

        :param application_id: The ID of the application.
        :return: Details of the application.
        """

        return self.britive.get(f'{self.base_url}/{application_id}')

    def catalog(self) -> list:
        """
        Return a list of applications in the application catalog.

        All applications in the tenant will derive from one of the applications returned.

        :return: List of applications in the application catalog
        """

        options = self.britive.get(f'{self.britive.base_url}/system/apps')
        return [o for o in options if o['catalogAppId'] != 0]

    def create(self, catalog_id: int, application_name: str) -> dict:
        """
        Create a new application.

        :param catalog_id: The ID of the catalog application from which to create the new application. Make a call to
            `catalog` to retrieve the list of applications in the catalog.
        :param application_name: The unique name for the application being created.
        :return: Details of the created application.
        """

        data = {
            'catalogAppId': catalog_id,
            'catalogAppDisplayName': application_name
        }

        return self.britive.post(self.base_url, json=data)

    def set_user_account_mapping(self, application_id: str, user_account_mapping: str = None) -> dict:
        """
        Set the user-to-account mapping attribute of the application.

        :param application_id: The ID of the application to update.
        :param user_account_mapping: The user-to-account mapping from Britive to the destination
            cloud provider identity. Valid values are `email`, `username`, and None.
        :return:
        """

        if user_account_mapping and user_account_mapping not in valid_user_account_mappings.keys():
            raise ValueError('invalid user_account_mapping value')

        data = {
            'userAccountMappings': valid_user_account_mappings.get(user_account_mapping, [])
        }
        self.britive.post(f'{self.base_url}/{application_id}/user-account-mappings', json=data)
        return self.get(application_id=application_id)

    def enable(self, application_id: str) -> dict:
        """
        Enable an application.

        :param application_id: The ID of the application to enable.
        :return: Details of the newly enabled application.
        """

        data = {
            'status': 'active'
        }
        return self.britive.patch(f'{self.base_url}/{application_id}', json=data)

    def disable(self, application_id: str) -> dict:
        """
        Disable an application.

        :param application_id: The ID of the application to disable.
        :return: Details of the newly disabled application.
        """

        data = {
            'status': 'inactive'
        }
        return self.britive.patch(f'{self.base_url}/{application_id}', json=data)

    def test(self, application_id: str) -> dict:
        """
        Test an application configuration.

        This operation performs configuration checks to ensure that the application is correctly configured.

        :param application_id: The ID of the application to test.
        :return: A dict containing 2 keys - `success` and `message`.
        """

        return self.britive.get(f'{self.base_url}/{application_id}/test')

    def update(self, application_id: str, **kwargs) -> dict:
        """
        Update an application.

        :param application_id: The ID of the application to update.
        :param kwargs: A set of name/value pairs where the property name will be updated to the supplied value.
            Property names are dependent upon the catalog application. Call method `applications.catalog()` for a list
            of all catalog applications, find the catalog application in question, and review the attribute
            `propertyTypes` for a list of acceptable names to use here.
        :return: Details of the updated application.
        """

        data = {
            'propertyTypes': []
        }
        for key, value in kwargs.items():
            data['propertyTypes'].append(
                {
                    'name': key,
                    'value': value,
                    'defaultValue': value
                }
            )
        return self.britive.patch(f'{self.base_url}/{application_id}/properties', json=data)

    def scan(self, application_id: str) -> dict:
        """
        Initiate a scan of the application.

        For applications which have `catalogApplication.supportsEnvironmentScanning` set to `False` do not scan the
        application. Instead scan the environment(s) associated with the application.

        For AWS (not AWS standalone) application types, this method will initiate an organization scan to collect
        the accounts in the org and their place in the OU structure.

        Scans are asynchronous operations. The response will include a `taskId` which can be used to make calls
        to `britive.scans.status()` to obtain the current status of the scan.

        Note that scans can also be initiated from the Scans class. The same type of scan will be performed no matter
        where it is initiated.

        :param application_id: The ID of the application to scan.
        :return: Details of the scan that was initiated.
        """

        return self.britive.scans.scan(application_id=application_id)

    def delete(self, application_id: str) -> None:
        """
        Delete an application.

        Will also recursively delete any environments, environment groups, accounts, profiles, etc. which are a part
        of the application.

        :param application_id: The ID of the application to delete.
        :return: None
        """

        params = {
            'appContainerId': application_id
        }
        return self.britive.delete(self.base_url, params=params)



class Permissions:
    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/apps'

    def list(self, application_id: str, environment_id: str = None,
             include_associations: bool = True, filter_expression: str = None) -> list:
        """
        Return details of all the permissions associated with a given application and environment.

        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param include_associations: Include the associated accounts and groups in the response.
        :param filter_expression: Filter based on `scanStatus`. Example: `scanStatus eq Unchanged`.
        :return: List containing details of each permission.
        """

        params = {
            'page': 0,
            'size': 100,
            'includeMembers': include_associations
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        if filter_expression:
            params['filter'] = filter_expression

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/permissions'
        return self.britive.get(url, params=params)

    def accounts(self, permission_id: str, application_id: str,
                 environment_id: str = None, filter_expression: str = None) -> list:
        """
        Return details about the accounts associated with the specified permission.

        :param permission_id: The ID of the permission.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param filter_expression: Filter based on `scanStatus`. Example: `scanStatus eq Unchanged`.
        :return: List of permissions associated with the specified account.
        """

        params = {
            'page': 0,
            'size': 100
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        if filter_expression:
            params['filter'] = filter_expression

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/permissions/{permission_id}/accounts'
        return self.britive.get(url, params=params)

    def groups(self, permission_id: str, application_id: str,
               environment_id: str = None, filter_expression: str = None) -> list:
        """
        Return details about the groups associated to the specified account.

        :param permission_id: The ID of the permission.
        :param application_id: The ID of the application.
        :param environment_id: Optionally the ID of the environment. Required only for applications which have
            `catalogApplication.supportsEnvironmentScanning` set to `False`.
        :param filter_expression: Filter based on `scanStatus`. Example: `scanStatus eq Unchanged`.
        :return: List of groups associated with the specified account.
        """

        params = {
            'page': 0,
            'size': 100
        }

        environment_id = environment_id or self.britive.get_root_environment_group(application_id)

        if filter_expression:
            params['filter'] = filter_expression

        url = f'{self.base_url}/{application_id}/environments/{environment_id}/permissions/{permission_id}/groups'
        return self.britive.get(url, params=params)



from datetime import datetime, timezone, timedelta
import time
from . import exceptions


class MySecrets:
    """
    This class is meant to be called by end users (as part of custom API integration work or pybritive CLI).
    It is an API layer on top of the actions that can be performed on the "My Secrets" page of the Britive UI.

    No "administrative" access is required by the methods in this class. Each method will only return resources/allow
    actions which are permitted to be performed by the user/service identity, as identified by an API token or
    interactive login bearer token.

    It is entirely possible that an administrator who makes these API calls could get nothing returned, as that
    administrator may not have access to any secrets.
    """

    def __init__(self, britive):
        self.britive = britive
        self.base_url = f'{self.britive.base_url}/v1/secretmanager'

    def __get_vault_id(self):
        # only 1 vault is allowed per tenant, so we can reliably grab the ID of that vault
        try:
            return self.britive.get(f'{self.base_url}/vault')['id']
        except KeyError as e:
            if 'id' in str(e):
                raise exceptions.NoSecretsVaultFound()

    def list(self, path: str = '/', search: str = None) -> list:
        """
        Recursively list all secrets under the given path.

        Only entries/secrets assigned to the provided identity, via a policy, are listed.

        :param path: Optional argument used to specify where in the hierarchy to begin listing secrets. Include
            leading / if provided.
        :param search: Optional argument used to filter the list of returned secrets by searching on the secret name.
        :return: List of secrets for which the provided identity has access to view.
        """

        params = {
            'recursiveSecrets': True,
            'getmetadata': True,
            'path': path,
            'type': 'secret'
        }

        if search:
            params['filter'] = f"name co '{search}'"

        return self.britive.get(f'{self.base_url}/vault/{self.__get_vault_id()}/secrets', params=params)

    def view(self, path: str, justification: str = None, wait_time: int = 60,
             max_wait_time: int = 600) -> dict:
        """
        Retrieve the decrypted secret value.

        :param path: The path to the secret. Include the leading /.
        :param justification: Optional justification if viewing the secret requires approval.
        :param wait_time: The number of seconds to sleep/wait between polling to check if the secret
            request was approved.
        :param max_wait_time: The maximum number of seconds to wait for an approval before throwing
            an exception.
        :return: Details of the decrypted secret.
        :raises ApprovalRequiredButNoJustificationProvided: if approval is required but no justification is provided.
        :raises AccessDenied: if the caller does not have access to the secret being requested.
        :raises ApprovalWorkflowTimedOut: if max_wait_time has been reached while waiting for approval.
        :raises ApprovalWorkflowRejected: if the request to view the secret was rejected.
        :raises InvalidRequest: if the node/secret does not exist
        """

        vault_id = self.__get_vault_id()
        quit_time = datetime.now(timezone.utc) + timedelta(seconds=max_wait_time)
        params = {
            'path': path
        }
        data = {
            'justification': justification
        }
        first = True
        while True:  # this is not loop forever due to exceptions raised or returning the secret value
            try:
                # handle when the time has expired waiting for approval
                if datetime.now(timezone.utc) >= quit_time:
                    raise exceptions.ApprovalWorkflowTimedOut()

                # attempt to get the secret value and return it
                return self.britive.post(
                    f'{self.base_url}/vault/{vault_id}/accesssecrets',
                    params=params,
                    json=data if first else None
                )['value']

            # 403 will be returned when approval is required or pending or access is denied
            except exceptions.ForbiddenRequest as e:
                if 'PE-0011' in str(e) and not justification:
                    if first:
                        raise exceptions.ApprovalRequiredButNoJustificationProvided()
                    else:
                        raise exceptions.ApprovalWorkflowRejected()
                if 'PE-0002' in str(e):
                    raise exceptions.AccessDenied()
                if 'PE-0010' in str(e):  # approval to view the secret is pending...
                    first = False
                    time.sleep(wait_time)
                else:
                    raise e

    def download(self, path: str, justification: str = None, wait_time: int = 60, max_wait_time: int = 600) -> dict:
        """
        Retrieve the decrypted secret file.

        It is left to the caller of this method to persist the file contents to disk. Example of usage is below.

        r = b.my_secrets.download(path='/examplefile')
        with open(r['filename'], 'wb') as f:
            f.write(r['content_bytes'])

        :param path: The path to the secret. Include the leading /.
        :param justification: Optional justification if viewing the secret requires approval.
        :param wait_time: The number of seconds to sleep/wait between polling to check if the secret
            request was approved.
        :param max_wait_time: The maximum number of seconds to wait for an approval before throwing
            an exception.
        :return: Dict containing the filename of the downloaded file and the content of the file as bytes.
        :raises ApprovalRequiredButNoJustificationProvided: if approval is required but no justification is provided.
        :raises AccessDenied: if the caller does not have access to the secret being requested.
        :raises ApprovalWorkflowTimedOut: if max_wait_time has been reached while waiting for approval.
        """

        vault_id = self.__get_vault_id()
        quit_time = datetime.now(timezone.utc) + timedelta(seconds=max_wait_time)
        params = {
            'path': path
        }

        try:
            # attempt to get the secret file and return it

            return self.britive.get(
                f'{self.base_url}/vault/{vault_id}/downloadfile',
                params=params
            )
        # 403 will be returned when approval is required or access is denied
        except exceptions.ForbiddenRequest as e:
            if 'PE-0011' in str(e):  # justification is required which means we have an approval workflow to deal with
                # lets call view so we can go through the full approval process
                self.view(
                    path=path,
                    justification=justification,
                    wait_time=wait_time,
                    max_wait_time=max_wait_time
                )

                # and then we can get the file again
                return self.britive.get(
                    f'{self.base_url}/vault/{vault_id}/downloadfile',
                    params=params
                )
            if 'PE-0002' in str(e):
                raise exceptions.AccessDenied()
            else:
                raise e



from .cache import *  # will also import some globals like `britive`


def test_create(cached_task):
    assert isinstance(cached_task, dict)
    assert cached_task['name'] == 'test'


def test_list(cached_task_service):
    tasks = britive.tasks.list(task_service_id=cached_task_service['taskServiceId'])
    assert isinstance(tasks, list)
    assert len(tasks) == 1
    assert tasks[0]['name'] == 'test'


def test_get(cached_task_service, cached_task):
    task = britive.tasks.get(task_service_id=cached_task_service['taskServiceId'], task_id=cached_task['taskId'])
    assert isinstance(task, dict)
    assert task['name'] == 'test'


def test_update(cached_task_service, cached_task):
    task = britive.tasks.update(
        task_service_id=cached_task_service['taskServiceId'],
        task_id=cached_task['taskId'],
        name='test2'
    )
    assert isinstance(task, dict)
    assert task['name'] == 'test2'


def test_delete(cached_task_service, cached_task):
    task = britive.tasks.delete(task_service_id=cached_task_service['taskServiceId'], task_id=cached_task['taskId'])
    assert task is None
    tasks = britive.tasks.list(task_service_id=cached_task_service['taskServiceId'])
    assert len(tasks) == 0
    cleanup('task')

